@@ -9,6 +9,7 @@ use warnings;
 
 use Net::UPnP;
 use Net::UPnP::ActionResponse;
+use Net::UPnP::QueryResponse;
 
 use vars qw($_DEVICE $_DEVICE_DESCRIPTION $SERVICETYPE $SERVICEID $SCPDURL $CONTROLURL $EVENTSUBURL);
 
@@ -61,6 +62,16 @@ sub setdevicedescription() {
 
 sub getdevicedescription() {
 	my($this) = shift;
+	my %args = (
+		name => undef,	
+		@_,
+	);
+	if ($args{name}) {
+		unless ($this->{$Net::UPnP::Service::_DEVICE_DESCRIPTION} =~ m/<$args{name}>(.*)<\/$args{name}>/i) {
+			return '';
+		}
+	 	return $1;
+	}
 	$this->{$Net::UPnP::Service::_DEVICE_DESCRIPTION};
  }
 
@@ -70,10 +81,7 @@ sub getdevicedescription() {
 
 sub getservicetype() {
 	my($this) = shift;
-	unless ($this->{$Net::UPnP::Service::_DEVICE_DESCRIPTION} =~ m/<$Net::UPnP::Service::SERVICETYPE>(.*)<\/$Net::UPnP::Service::SERVICETYPE>/i) {
-		return '';
-	}
- 	return $1;
+	$this->getdevicedescription(name => $Net::UPnP::Service::SERVICETYPE);
  }
 
 #------------------------------
@@ -82,10 +90,7 @@ sub getservicetype() {
 
 sub getserviceid() {
 	my($this) = shift;
-	unless ($this->{$Net::UPnP::Service::_DEVICE_DESCRIPTION} =~ m/<$Net::UPnP::Service::SERVICEID>(.*)<\/$Net::UPnP::Service::SERVICEID>/i) {
-		return '';
-	}
- 	return $1;
+	$this->getdevicedescription(name => $Net::UPnP::Service::SERVICEID);
  }
 
 #------------------------------
@@ -94,10 +99,7 @@ sub getserviceid() {
 
 sub getscpdurl() {
 	my($this) = shift;
-	unless ($this->{$Net::UPnP::Service::_DEVICE_DESCRIPTION} =~ m/<$Net::UPnP::Service::SCPDURL>(.*)<\/$Net::UPnP::Service::SCPDURL>/i) {
-		return '';
-	}
- 	return $1;
+	$this->getdevicedescription(name => $Net::UPnP::Service::SCPDURL);
  }
 
 #------------------------------
@@ -106,35 +108,75 @@ sub getscpdurl() {
 
 sub getcontrolurl() {
 	my($this) = shift;
-	unless ($this->{$Net::UPnP::Service::_DEVICE_DESCRIPTION} =~ m/<$Net::UPnP::Service::CONTROLURL>(.*)<\/$Net::UPnP::Service::CONTROLURL>/i) {
-		return '';
-	}
- 	return $1;
+	$this->getdevicedescription(name => $Net::UPnP::Service::CONTROLURL);
  }
 
 #------------------------------
-# geteventsubrul
+# geteventsuburl
 #------------------------------
 
-sub geteventsubrul() {
+sub geteventsuburl() {
 	my($this) = shift;
-	unless ($this->{$Net::UPnP::Service::_DEVICE_DESCRIPTION} =~ m/<$Net::UPnP::Service::EVENTSUBURL>(.*)<\/$Net::UPnP::Service::EVENTSUBURL>/i) {
-		return '';
-	}
- 	return $1;
+	$this->getdevicedescription(name => $Net::UPnP::Service::EVENTSUBURL);
  }
 
 #------------------------------
-# geteventsubrul
+# getposturl
 #------------------------------
 
-sub postcontrol() {
+sub getposturl() {
 	my($this) = shift;
-	my ($action_name, $action_arg) = @_;
+	my ($ctrl_url) = @_;
 	my (
 		$dev,
 		$location_url,
 		$url_base,
+	);
+	
+	$dev = $this->getdevice();
+	
+	$location_url = $dev->getlocation();
+	$url_base = $dev->geturlbase();
+	$ctrl_url = $this->getcontrolurl();
+
+	#print "$location_url\n";
+	#print "$url_base\n";
+	#print "$ctrl_url\n";
+		
+	unless ($ctrl_url =~ m/http:\/\/(.*)/i) {
+		if (0 < length($url_base)) {
+			# Thanks for Thus0 (2005/01/12)
+			if (rindex($url_base, '/') == (length($url_base)-1) && index($ctrl_url, '/') == 0) {
+				$ctrl_url = $url_base . substr($ctrl_url, 1);
+			} else {
+				$ctrl_url = $url_base . $ctrl_url;
+			}
+		}
+		else {
+			if ($location_url =~ m/http:\/\/([0-9a-z.]+)[:]*([0-9]*)\/(.*)/i) {
+				if (defined($3) && 0 < length($3)) {
+					$ctrl_url = "http:\/\/" . $1 . ":" . $2 . $ctrl_url;
+				} else {
+					$ctrl_url = "http:\/\/" . $1 . ":" . $2 . "\/" . $ctrl_url;
+				}
+			} else {
+				$ctrl_url = $location_url .  $ctrl_url;
+			}
+		}
+	}
+	
+	return $ctrl_url;
+}
+
+#------------------------------
+# postaction
+#------------------------------
+
+sub postaction() {
+	my($this) = shift;
+	my ($action_name, $action_arg) = @_;
+	my (
+		$dev,
 		$ctrl_url,
 		$service_type,
 		$soap_action,
@@ -153,33 +195,8 @@ sub postcontrol() {
 	$action_res = Net::UPnP::ActionResponse->new();
 	
 	$dev = $this->getdevice();
-	
-	$location_url = $dev->getlocation();
-	$url_base = $dev->geturlbase();
 	$ctrl_url = $this->getcontrolurl();
-
-	#print "$location_url\n";
-	#print "$url_base\n";
-	#print "$ctrl_url\n";
-		
-	unless ($ctrl_url =~ m/http:\/\/(.*)/i) {
-		if (0 < length($url_base)) {
-			$ctrl_url = $url_base . $ctrl_url;
-		}
-		else {
-			if ($location_url =~ m/http:\/\/([0-9a-z.]+)[:]*([0-9]*)\/(.*)/i) {
-				if (defined($3) && 0 < length($3)) {
-					$ctrl_url = "http:\/\/" . $1 . ":" . $2 . $ctrl_url;
-				} else {
-					$ctrl_url = "http:\/\/" . $1 . ":" . $2 . "\/" . $ctrl_url;
-				}
-			} else {
-				$ctrl_url = $location_url .  $ctrl_url;
-			}
-		}
-	}
-
-	#print "$ctrl_url\n";
+	$ctrl_url = $this->getposturl($ctrl_url);
 	
 	unless ($ctrl_url =~ m/http:\/\/([0-9a-z.]+)[:]*([0-9]*)\/(.*)/i) {
 		#print "Invalid URL : $ctrl_url\n";
@@ -189,9 +206,13 @@ sub postcontrol() {
 	}
 	$post_addr = $1;
 	$post_port = $2;
-	$post_path = "\/" . $3;
-	#$post_path = $this->getcontrolurl();
-	
+	if (index($3, '/') == 0) {
+		$post_path = $3;
+	}
+	else {
+		$post_path = "\/" . $3;
+	}
+
 	$service_type = $this->getservicetype();
 	$soap_action = "\"" . $service_type . "#" . $action_name . "\"";
 
@@ -227,6 +248,80 @@ SOAP_CONTENT
 	return $action_res;
 }
 
+#------------------------------
+# postcontrol
+#------------------------------
+
+sub postcontrol() {
+	my($this) = shift;
+	my ($action_name, $action_arg) = @_;
+	return $this->postaction($action_name, $action_arg);
+}
+
+#------------------------------
+# postquery
+#------------------------------
+
+sub postquery() {
+	my($this) = shift;
+	my ($var_name) = @_;
+	my (
+		$dev,
+		$ctrl_url,
+		$service_type,
+		$soap_action,
+		$soap_content,
+		$post_addr,
+		$post_port,
+		$post_path,
+		$http_req,
+		$post_res,
+		$query_res,
+	);
+	
+	$query_res = Net::UPnP::QueryResponse->new();
+	
+	$dev = $this->getdevice();
+	$ctrl_url = $this->getcontrolurl();
+	$ctrl_url = $this->getposturl($ctrl_url);
+	
+	unless ($ctrl_url =~ m/http:\/\/([0-9a-z.]+)[:]*([0-9]*)\/(.*)/i) {
+		#print "Invalid URL : $ctrl_url\n";
+		$post_res = Net::UPnP::HTTPResponse->new();
+		$query_res->sethttpresponse($post_res);
+		return $query_res;
+	}
+	$post_addr = $1;
+	$post_port = $2;
+	if (index($3, '/') == 0) {
+		$post_path = $3;
+	}
+	else {
+		$post_path = "\/" . $3;
+	}
+	
+	$service_type = $this->getservicetype();
+	$soap_action = "\"urn:schemas-upnp-org:control-1-0#QueryStateVariable\"";
+
+$soap_content = <<"SOAP_CONTENT";
+<?xml version=\"1.0\" encoding=\"utf-8\"?>
+<s:Envelope xmlns:s=\"http:\/\/schemas.xmlsoap.org\/soap\/envelope\/\" s:encodingStyle=\"http:\/\/schemas.xmlsoap.org\/soap\/encoding/\">
+\t<s:Body>
+\t\t<u:QueryStateVariable xmlns:u=\"urn:schemas-upnp-org:control-1-0\">
+\t\t\t<u:varName>$var_name</u:varName>
+\t\t</u:QueryStateVariable>
+\t</s:Body>
+</s:Envelope>
+SOAP_CONTENT
+
+	$http_req = Net::UPnP::HTTP->new();
+	$post_res = $http_req->postsoap($post_addr, $post_port, $post_path, $soap_action, $soap_content);
+
+	$query_res->sethttpresponse($post_res);
+	
+	return $query_res;
+}
+
 1;
 
 __END__
@@ -294,6 +389,16 @@ The package is used a object of UPnP service.
 
 Get the parent device of the service.
 
+=item B<getdevicedescription> - get the service description of the device description.
+
+    $description = $service->getdevicedescription(
+    	                        name => $name # undef
+                             );
+
+Get the service description of the device description. 
+	
+The function returns the all description when the name parameter is not specified, otherwise return a value the specified name.
+
 =item B<getservicetype> - get the service type.
 
     $service_type = $service->getservicetype();
@@ -306,11 +411,19 @@ Get the service type.
 
 Get the service id.
 
-=item B<postcontrol> - post a control action.
+=item B<postaction> - post a action control.
 
     $action_res = $service->postcontrol($action_name, \%action_arg);
 
-Post a control action to the device, and return L<Net::UPnP::ActionResponse>.
+Post a action control to the device, and return L<Net::UPnP::ActionResponse>.
+
+The method was renamed from postcontrol(), but the old name is deprecated.
+	
+=item B<postquery> - post a query control.
+
+    $query_res = $service->postcontrol($var_name);
+
+Post a query control to the device, and return L<Net::UPnP::QueryResponse>.
 
 =back
 
